using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace tonegen
{
    class Program
    {
        enum Quality
        {
            Point,
            Linear,
            Cubic
        };

        static void Main(string[] args)
        {
            if (args.Length < 2 || args.Length > 3)
            {
                Console.WriteLine("usage: wavtable <infile> <outfile> <optional sourcelen>");
                Console.WriteLine("hit a key to continue...");
                Console.ReadKey();
                return;
            }

            BinaryReader reader = new BinaryReader(File.Open(args[0], FileMode.Open));
            BinaryWriter writer = new BinaryWriter(File.Open(args[1], FileMode.Create));

            long sourceLen = 0;
            if (args.Length == 3)
            {
                sourceLen = System.Convert.ToInt32(args[2]);
            }
            else
            {
                sourceLen = reader.BaseStream.Length/2;
            }

            Console.WriteLine("sourceLen: " + sourceLen);

            short[] sourceTable = new short[sourceLen * 2];

            for (int i=0; i<sourceLen; i++)
            {
                sourceTable[i] = reader.ReadInt16();
                sourceTable[i + sourceLen] = sourceTable[i];
            }
            /*
            WriteTable(writer, sourceTable, sourceLen, 3200, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 1600, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 800, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 400, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 200, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 100, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 50, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 25, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 12, Quality.Point);
            WriteTable(writer, sourceTable, sourceLen, 6, Quality.Point);

            WriteTable(writer, sourceTable, sourceLen, 3200, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 1600, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 800, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 400, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 200, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 100, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 50, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 25, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 12, Quality.Linear);
            WriteTable(writer, sourceTable, sourceLen, 6, Quality.Linear);
            */
            WriteTable(writer, sourceTable, sourceLen, 6400, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 3200, Quality.Cubic);       
            WriteTable(writer, sourceTable, sourceLen, 1600, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 800, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 400, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 200, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 100, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 50, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 25, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 12, Quality.Cubic);
            WriteTable(writer, sourceTable, sourceLen, 6, Quality.Cubic);
           
            //WriteTable(writer, sourceTable, sourceLen, 3200, Quality.Linear);
            //WriteTable(writer, sourceTable, sourceLen, 3200, Quality.Cubic);

            writer.Close();
            reader.Close();

            Console.WriteLine("hit a key to continue...");
            Console.ReadKey();
        }

        static void WriteTable(BinaryWriter writer, short[] sourceTable, long sourceLen, int destSize, Quality quality)
        {
            
            // we are resampling down -- this is very useful for very small values - like when destSize is 6
            // need to tweak it to be more accurate at larger sizes
            if (destSize < sourceLen) 
            //if (destSize < 100)
            {
                // TODO: figure out why its not handling it right for destSize 3200
                double chunkSize = (double)sourceLen / (double)destSize;
                long half1 = (long)chunkSize / 2;
                long half2 = (long)chunkSize - half1;
                for (int i = 0; i < destSize; i++)
                {
                    long chunkSum = 0;
                    long offset = (long)((double)i * chunkSize);
                    for (int c = 0; c < half1; c++)
                    {
                        chunkSum += sourceTable[offset + c - half1 + sourceLen];
                    }

                    for (int c = 0; c < half2; c++)
                    {
                        chunkSum += sourceTable[offset + c];
                    }

                    short value = (short)(chunkSum / (long)chunkSize);
                    writer.Write(value);
                }
                
                return;
            }
            

            // we are sampling up
            for (int i = 0; i < destSize; i++)
            {
                // quick point sampling
                double fScaled = (double)i * ((double)sourceLen / (double)destSize);
                long scaledIndex = (long)fScaled;
                /*
                if (scaledIndex > sourceLen)
                {
                    Console.WriteLine("error got scaledIndex:" + scaledIndex + " i: " + i);
                }
                 */
                short value = 0;
                if (quality == Quality.Point)
                {
                    value = sourceTable[scaledIndex];
                }
                else if (quality == Quality.Linear)
                {
                    double floatingPart = fScaled - scaledIndex;
                    long scaledIndex2 = scaledIndex + 1;

                    double fValue = (floatingPart * sourceTable[scaledIndex]) + ((1.0 - floatingPart) * sourceTable[scaledIndex2]);
                    value = (short)fValue;
                }
                else if (quality == Quality.Cubic)
                {
                    //cubic interpolation start		

					float floatingPart = (float)(fScaled - (long)(scaledIndex));

					double [] xarray = new double[4]; //0 = xm1, 1 = x0, 2 = x1, 3 = x2

                    xarray[0] = sourceTable[scaledIndex + sourceLen - 1];
                    xarray[1] = sourceTable[scaledIndex];
                    xarray[2] = sourceTable[scaledIndex + 1];
                    xarray[3] = sourceTable[scaledIndex + 2];

					double a = (3 * (xarray[1]-xarray[2]) - xarray[0] + xarray[3]) / 2;
					double b = 2*xarray[2] + xarray[0] - (5*xarray[1] + xarray[2]) / 2;
				 	double c = (xarray[2] - xarray[0]) / 2;
					double fValue = (((a * floatingPart) + b) * floatingPart + c) * floatingPart + xarray[1];
                    value = (short)fValue;
                }
                //Console.WriteLine("i:" + i + " scaledIndex: " + scaledIndex);
                writer.Write(value);
            }

            Console.WriteLine("sourceLen:" + sourceLen + " destSize: " + destSize);
        }
    }
}
